@@ -1,5 +1,5 @@
 # emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-; coding: utf-8 -*-
-# ex: set sts=4 ts=4 sw=4 noet:
+# ex: set sts=4 ts=4 sw=4 et:
 # ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
 #
 #   See COPYING file distributed along with the datalad package for the
@@ -14,64 +14,64 @@ Note: Tests of `run` that involve `rerun` are in interface.tests.test_run.
 __docformat__ = 'restructuredtext'
 
 import logging
-
 import os
 import os.path as op
+import sys
 from os import (
     mkdir,
     remove,
 )
-import sys
-
 from unittest.mock import patch
 
+from datalad.api import (
+    clone,
+    run,
+)
+from datalad.cli.main import main
+from datalad.core.local.run import (
+    _format_iospecs,
+    _get_substitutions,
+    format_command,
+    run_command,
+)
+from datalad.distribution.dataset import Dataset
+from datalad.support.exceptions import (
+    CommandError,
+    IncompleteResultsError,
+    NoDatasetFound,
+)
+from datalad.tests.utils import (
+    DEFAULT_BRANCH,
+    OBSCURE_FILENAME,
+    assert_false,
+    assert_in,
+    assert_in_results,
+    assert_not_in,
+    assert_not_in_results,
+    assert_raises,
+    assert_repo_status,
+    assert_result_count,
+    assert_status,
+    create_tree,
+    eq_,
+    known_failure_windows,
+    neq_,
+    ok_,
+    ok_exists,
+    ok_file_has_content,
+    patch_config,
+    swallow_logs,
+    swallow_outputs,
+    with_tempfile,
+    with_tree,
+)
 from datalad.utils import (
     chpwd,
     ensure_unicode,
     on_windows,
 )
 
-from datalad.cmdline.main import main
-from datalad.distribution.dataset import Dataset
-from datalad.support.exceptions import (
-    CommandError,
-    NoDatasetFound,
-    IncompleteResultsError,
-)
-from datalad.api import (
-    run,
-)
-from datalad.core.local.run import (
-    format_command,
-    run_command,
-)
-from datalad.tests.utils import (
-    assert_false,
-    assert_in,
-    assert_in_results,
-    assert_not_in,
-    assert_raises,
-    assert_repo_status,
-    assert_result_count,
-    assert_status,
-    create_tree,
-    DEFAULT_BRANCH,
-    eq_,
-    known_failure_appveyor,
-    known_failure_githubci_win,
-    known_failure_windows,
-    neq_,
-    OBSCURE_FILENAME,
-    ok_,
-    ok_exists,
-    ok_file_has_content,
-    slow,
-    swallow_logs,
-    swallow_outputs,
-    with_tempfile,
-    with_testrepos,
-    with_tree,
-)
+cat_command = 'cat' if not on_windows else 'type'
 
 
 @with_tempfile(mkdir=True)
@@ -100,15 +100,17 @@ def test_basics(path, nodspath):
     with chpwd(path), \
             swallow_outputs():
         # provoke command failure
-        with assert_raises(CommandError) as cme:
-            ds.run('7i3amhmuch9invalid')
+        res = ds.run('7i3amhmuch9invalid', on_failure="ignore",
+                     result_renderer=None)
+        assert_result_count(res, 1, action="run", status="error")
+        run_res = [r for r in res if r["action"] == "run"][0]
         # let's not speculate that the exit code is always 127
-        ok_(cme.exception.code > 0)
+        ok_(run_res["run_info"]["exit"] > 0)
         eq_(last_state, ds.repo.get_hexsha())
         # now one that must work
         res = ds.run('cd .> empty', message='TEST')
         assert_repo_status(ds.path)
-        assert_result_count(res, 2)
+        assert_result_count(res, 3)
         # TODO 'state' is still untracked!!!
         assert_result_count(res, 1, action='add',
                             path=op.join(ds.path, 'empty'), type='file')
@@ -143,6 +145,33 @@ def test_basics(path, nodspath):
             ds.run()
             assert_in("No command given", cml.out)
 
+    # running without a command is a noop
+    with chpwd(path):
+        with swallow_logs(new_level=logging.INFO) as cml:
+            assert_raises(
+                IncompleteResultsError,
+                ds.run,
+                '7i3amhmuch9invalid',
+                # this is on_failure=stop by default
+            )
+            # must give recovery hint in Python notation
+            assert_in("can save the changes with \"Dataset(", cml.out)
+
+    with chpwd(path):
+        # make sure that an invalid input declaration prevents command
+        # execution by default
+        assert_raises(
+            IncompleteResultsError,
+            ds.run, 'cd .> dummy0', inputs=['not-here'])
+        ok_(not (ds.pathobj / 'dummy0').exists())
+        # but the default behavior can be changed
+        assert_raises(
+            IncompleteResultsError,
+            ds.run, 'cd .> dummy0', inputs=['not-here'],
+            on_failure='continue')
+        # it has stilled failed, but the command got executed nevertheless
+        ok_((ds.pathobj / 'dummy0').exists())
+
 
 @known_failure_windows
 # ^ For an unknown reason, appveyor started failing after we removed
@@ -176,42 +205,36 @@ def test_py2_unicode_command(path):
         assert_repo_status(ds.path)
         ok_exists(op.join(path, u" β1 "))
 
-    with assert_raises(CommandError), swallow_outputs():
-        ds.run(u"bβ2.dat")
+    assert_in_results(
+        ds.run(u"bβ2.dat", result_renderer=None, on_failure="ignore"),
+        status="error", action="run")
 
 
 @with_tempfile(mkdir=True)
 def test_sidecar(path):
     ds = Dataset(path).create()
-    if ds.repo.is_managed_branch():
-        if "8" < ds.repo.git_annex_version < "8.20200309":
-            # git-annex's 1978a2420 (2020-03-09) fixed a bug where
-            # annexed dotfiles could switch when annex.dotfiles=true
-            # was not set in .git/config or git-annex:config.log.
-            ds.repo.config.set("annex.dotfiles", "true",
-                               where="local", reload=True)
     # Simple sidecar message checks.
     ds.run("cd .> dummy0", message="sidecar arg", sidecar=True)
     assert_not_in('"cmd":', ds.repo.format_commit("%B"))
 
-    ds.config.set("datalad.run.record-sidecar", "false", where="local")
+    ds.config.set("datalad.run.record-sidecar", "false", scope="local")
     ds.run("cd .> dummy1", message="sidecar config")
 
     assert_in('"cmd":', last_commit_msg(ds.repo))
 
-    ds.config.set("datalad.run.record-sidecar", "true", where="local")
+    ds.config.set("datalad.run.record-sidecar", "true", scope="local")
     ds.run("cd .> dummy2", message="sidecar config")
     assert_not_in('"cmd":', last_commit_msg(ds.repo))
 
     # Don't break when config.get() returns multiple values. Here it's two
     # values in .gitconfig, but a more realistic scenario is a value in
     # $repo/.git/config that overrides a setting in ~/.config/git/config.
-    ds.config.add("datalad.run.record-sidecar", "false", where="local")
+    ds.config.add("datalad.run.record-sidecar", "false", scope="local")
     ds.run("cd .> dummy3", message="sidecar config")
     assert_in('"cmd":', last_commit_msg(ds.repo))
 
 
-    # make sure sidecar file is committed when explicitly specifiying outputs
+    # make sure sidecar file is committed when explicitly specifying outputs
     ds.run("cd .> dummy4",
            outputs=["dummy4"],
            sidecar=True,
@@ -231,7 +254,6 @@ def test_run_save_deletion(path):
     assert_repo_status(ds.path)
 
 
-@known_failure_appveyor  # causes appveyor (only) to crash, reason unknown
 @with_tempfile(mkdir=True)
 def test_run_from_subds(path):
     subds = Dataset(path).create().create("sub")
@@ -277,12 +299,82 @@ def test_run_from_subds_gh3551(path):
         ok_(subds.repo.file_has_content("f"))
 
 
-# unexpected content of state "modified", likely a more fundamental issue with the
-# testrepo setup
-@known_failure_githubci_win
-@with_testrepos('basic_annex', flavors=['clone'])
-def test_run_explicit(path):
-    ds = Dataset(path)
+@with_tempfile(mkdir=True)
+def test_run_assume_ready(path):
+    ds = Dataset(path).create()
+    repo = ds.repo
+    adjusted = repo.is_managed_branch()
+
+    # --assume-ready=inputs
+
+    (repo.pathobj / "f1").write_text("f1")
+    ds.save()
+
+    def cat_cmd(fname):
+        return [sys.executable, "-c",
+                "import sys; print(open(sys.argv[-1]).read())",
+                fname]
+
+    assert_in_results(
+        ds.run(cat_cmd("f1"), inputs=["f1"]),
+        action="get", type="file")
+    # Same thing, but without the get() call.
+    assert_not_in_results(
+        ds.run(cat_cmd("f1"), inputs=["f1"], assume_ready="inputs"),
+        action="get", type="file")
+
+    ds.drop("f1", check=False)
+    if not adjusted:
+        # If the input is not actually ready, the command will fail.
+        assert_in_results(
+            ds.run(cat_cmd("f1"), inputs=["f1"], assume_ready="inputs",
+                   on_failure="ignore", result_renderer=None),
+            action="run", status="error")
+
+    # --assume-ready=outputs
+
+    def unlink_and_write_cmd(fname):
+        # This command doesn't care whether the output file is unlocked because
+        # it removes it ahead of time anyway.
+        return [sys.executable, "-c",
+                "import sys; import os; import os.path as op; "
+                "f = sys.argv[-1]; op.lexists(f) and os.unlink(f); "
+                "open(f, mode='w').write(str(sys.argv))",
+                fname]
+
+    (repo.pathobj / "f2").write_text("f2")
+    ds.save()
+
+    res = ds.run(unlink_and_write_cmd("f2"), outputs=["f2"])
+    if not adjusted:
+        assert_in_results(res, action="unlock", type="file")
+    # Same thing, but without the unlock() call.
+    res = ds.run(unlink_and_write_cmd("f2"), outputs=["f2"],
+                 assume_ready="outputs")
+    assert_not_in_results(res, action="unlock", type="file")
+
+    # --assume-ready=both
+
+    res = ds.run(unlink_and_write_cmd("f2"),
+                 outputs=["f2"], inputs=["f2"])
+    assert_in_results(res, action="get", type="file")
+    if not adjusted:
+        assert_in_results(res, action="unlock", type="file")
+
+    res = ds.run(unlink_and_write_cmd("f2"),
+                 outputs=["f2"], inputs=["f2"],
+                 assume_ready="both")
+    assert_not_in_results(res, action="get", type="file")
+    assert_not_in_results(res, action="unlock", type="file")
+
+
+@with_tempfile()
+@with_tempfile()
+def test_run_explicit(origpath, path):
+    origds = Dataset(origpath).create()
+    (origds.pathobj / "test-annex.dat").write_text('content')
+    origds.save()
+    ds = clone(origpath, path)
 
     assert_false(ds.repo.file_has_content("test-annex.dat"))
 
@@ -293,16 +385,18 @@ def test_run_explicit(path):
         ofh.write(", more")
 
     # We need explicit=True to run with dirty repo.
-    assert_status("impossible",
-                  ds.run("cat test-annex.dat test-annex.dat >doubled.dat",
-                         inputs=["test-annex.dat"],
-                         on_failure="ignore"))
+    assert_status(
+        "impossible",
+        ds.run(f"{cat_command} test-annex.dat test-annex.dat >doubled.dat",
+               inputs=["test-annex.dat"],
+               on_failure="ignore"))
 
     hexsha_initial = ds.repo.get_hexsha()
     # If we specify test-annex.dat as an input, it will be retrieved before the
     # run.
-    ds.run("cat test-annex.dat test-annex.dat >doubled.dat",
-           inputs=["test-annex.dat"], explicit=True)
+    ds.run(f"{cat_command} test-annex.dat test-annex.dat >doubled.dat",
+           inputs=["test-annex.dat"], explicit=True,
+           result_renderer='disabled')
     ok_(ds.repo.file_has_content("test-annex.dat"))
     # We didn't commit anything because outputs weren't specified.
     assert_false(ds.repo.file_has_content("doubled.dat"))
@@ -311,27 +405,28 @@ def test_run_explicit(path):
     # If an input doesn't exist, we just show the standard warning.
     with assert_raises(IncompleteResultsError):
         ds.run("ls", inputs=["not-there"], explicit=True,
-               on_failure="stop")
+               on_failure="stop", result_renderer='disabled')
 
     remove(op.join(path, "doubled.dat"))
 
     hexsha_initial = ds.repo.get_hexsha()
-    ds.run("cat test-annex.dat test-annex.dat >doubled.dat",
+    ds.run(f"{cat_command} test-annex.dat test-annex.dat >doubled.dat",
            inputs=["test-annex.dat"], outputs=["doubled.dat"],
-           explicit=True)
+           explicit=True, result_renderer='disabled')
     ok_(ds.repo.file_has_content("doubled.dat"))
-    assert_repo_status(ds.path, modified=["dirt_modified"], untracked=['dirt_untracked'])
+    assert_repo_status(ds.path, modified=["dirt_modified"],
+                       untracked=['dirt_untracked'])
     neq_(hexsha_initial, ds.repo.get_hexsha())
 
     # Saving explicit outputs works from subdirectories.
     subdir = op.join(path, "subdir")
     mkdir(subdir)
     with chpwd(subdir):
-        run("echo insubdir >foo", explicit=True, outputs=["foo"])
+        run("echo insubdir >foo", explicit=True, outputs=["foo"],
+            result_renderer='disabled')
     ok_(ds.repo.file_has_content(op.join("subdir", "foo")))
 
 
-@known_failure_windows  # due to use of obscure filename that breaks the runner on Win
 @with_tree(tree={OBSCURE_FILENAME + u".t": "obscure",
                  "bar.txt": "b",
                  "foo blah.txt": "f"})
@@ -403,30 +498,19 @@ def test_run_cmdline_disambiguation(path):
                 main(["datalad", "run", "--", "--message"])
             exec_cmd.assert_called_once_with(
                 '"--message"' if on_windows else "--message",
-                path, expected_exit=None)
+                path)
 
-        # And a twist on above: Our parser mishandles --version (gh-3067),
+        # Our parser used to mishandle --version (gh-3067),
         # treating 'datalad run CMD --version' as 'datalad --version'.
-        version_stream = "out"
-        with swallow_outputs() as cmo:
-            with assert_raises(SystemExit) as cm:
-                main(["datalad", "run", "echo", "--version"])
-            eq_(cm.exception.code, 0)
-            out = getattr(cmo, version_stream)
-        with swallow_outputs() as cmo:
-            with assert_raises(SystemExit):
-                main(["datalad", "--version"])
-            version_out = getattr(cmo, version_stream)
-        ok_(version_out)
-        eq_(version_out, out)
-        # We can work around that (i.e., make "--version" get passed as
-        # command) with "--".
-        with patch("datalad.core.local.run._execute_command") as exec_cmd:
-            with assert_raises(SystemExit):
-                main(["datalad", "run", "--", "echo", "--version"])
-            exec_cmd.assert_called_once_with(
-                '"echo" "--version"' if on_windows else "echo --version",
-                path, expected_exit=None)
+        # but that is no longer the case and echo --version should work with or
+        # without explicit "--" separator
+        for sep in [[], ['--']]:
+            with patch("datalad.core.local.run._execute_command") as exec_cmd:
+                with assert_raises(SystemExit):
+                    main(["datalad", "run"] + sep + ["echo", "--version"])
+                exec_cmd.assert_called_once_with(
+                    '"echo" "--version"' if on_windows else "echo --version",
+                    path)
 
 
 @with_tempfile(mkdir=True)
@@ -484,6 +568,36 @@ def test_run_path_semantics(path):
     assert_repo_status(ds0.path)
 
 
+@with_tempfile(mkdir=True)
+def test_run_remove_keeps_leading_directory(path):
+    ds = Dataset(op.join(path, "ds")).create()
+    repo = ds.repo
+
+    (ds.pathobj / "d").mkdir()
+    output = (ds.pathobj / "d" / "foo")
+    output.write_text("foo")
+    ds.save()
+
+    output_rel = str(output.relative_to(ds.pathobj))
+    repo.drop(output_rel, options=["--force"])
+
+    assert_in_results(
+        ds.run("cd .> {}".format(output_rel), outputs=[output_rel],
+               result_renderer='disabled'),
+        action="run.remove", status="ok")
+
+    assert_repo_status(ds.path)
+
+    # Remove still gets saved() if command doesn't generate the output (just as
+    # it would if git-rm were used instead of unlink).
+    repo.drop(output_rel, options=["--force"])
+    assert_in_results(
+        ds.run("cd .> something-else", outputs=[output_rel],
+               result_renderer='disabled'),
+        action="run.remove", status="ok")
+    assert_repo_status(ds.path)
+
+
 @with_tempfile(mkdir=True)
 def test_run_reglob_outputs(path):
     ds = Dataset(path).create()
@@ -541,3 +655,129 @@ def test_run_empty_repo(path):
     assert_status("ok", ds.run(cmd, inputs=["."]))
     assert_repo_status(ds.path)
     ok_exists(str(ds.pathobj / "foo"))
+
+
+@with_tree(tree={"foo": "f", "bar": "b"})
+def test_dry_run(path):
+    ds = Dataset(path).create(force=True)
+
+    # The dataset is reported as dirty, and the custom result render relays
+    # that to the default renderer.
+    with swallow_outputs() as cmo:
+        with assert_raises(IncompleteResultsError):
+            ds.run("blah ", dry_run="basic")
+        assert_in("run(impossible)", cmo.out)
+        assert_not_in("blah", cmo.out)
+
+    ds.save()
+
+    # unknown dry-run mode
+    assert_raises(ValueError, ds.run, 'blah', dry_run='absurd')
+
+    with swallow_outputs() as cmo:
+        ds.run("blah ", dry_run="basic")
+        assert_in("Dry run", cmo.out)
+        assert_in("location", cmo.out)
+        assert_in("blah", cmo.out)
+        assert_not_in("expanded inputs", cmo.out)
+        assert_not_in("expanded outputs", cmo.out)
+
+    with swallow_outputs() as cmo:
+        ds.run("blah {inputs} {outputs}", dry_run="basic",
+               inputs=["fo*"], outputs=["b*r"])
+        assert_in(
+            'blah "foo" "bar"' if on_windows else "blah foo bar",
+            cmo.out)
+        assert_in("expanded inputs", cmo.out)
+        assert_in("['foo']", cmo.out)
+        assert_in("expanded outputs", cmo.out)
+        assert_in("['bar']", cmo.out)
+
+    # Just the command.
+    with swallow_outputs() as cmo:
+        ds.run("blah ", dry_run="command")
+        assert_not_in("Dry run", cmo.out)
+        assert_in("blah", cmo.out)
+        assert_not_in("inputs", cmo.out)
+
+    # The output file wasn't unlocked.
+    assert_repo_status(ds.path)
+
+    # Subdaset handling
+
+    subds = ds.create("sub")
+    (subds.pathobj / "baz").write_text("z")
+    ds.save(recursive=True)
+
+    # If a subdataset is installed, it works as usual.
+    with swallow_outputs() as cmo:
+        ds.run("blah {inputs}", dry_run="basic", inputs=["sub/b*"])
+        assert_in(
+            'blah "sub\\baz"' if on_windows else 'blah sub/baz',
+            cmo.out)
+
+    # However, a dry run will not do the install/reglob procedure.
+    ds.drop("sub", what='all', reckless='kill', recursive=True)
+    with swallow_outputs() as cmo:
+        ds.run("blah {inputs}", dry_run="basic", inputs=["sub/b*"])
+        assert_in("sub/b*", cmo.out)
+        assert_not_in("baz", cmo.out)
+
+
+@with_tree(tree={OBSCURE_FILENAME + ".t": "obscure",
+                 "normal.txt": "normal"})
+def test_io_substitution(path):
+    files = [OBSCURE_FILENAME + ".t", "normal.txt"]
+    ds = Dataset(path).create(force=True)
+    ds.save()
+    # prefix the content of any given file with 'mod::'
+    cmd = "import sys; from pathlib import Path; t = [(Path(p), 'mod::' + Path(p).read_text()) for p in sys.argv[1:]]; [k.write_text(v) for k, v in t]"
+    cmd_str = "{} -c \"{}\" {{inputs}}".format(sys.executable, cmd)
+    # this should run and not crash with permission denied
+    ds.run(cmd_str, inputs=["{outputs}"], outputs=["*.t*"],
+           result_renderer='disabled')
+    # all filecontent got the prefix
+    for f in files:
+        ok_((ds.pathobj / f).read_text().startswith('mod::'))
+
+    # we could just ds.rerun() now, and it should work, but this would make
+    # rerun be a dependency of a core test
+    # instead just double-run, but with a non-list input-spec.
+    # should have same outcome
+    ds.run(cmd_str, inputs="{outputs}", outputs="*.t*",
+           result_renderer='disabled')
+    for f in files:
+        ok_((ds.pathobj / f).read_text().startswith('mod::mod::'))
+
+
+def test_format_iospecs():
+    seq = ['one', 'two']
+    eq_(seq, _format_iospecs(['{dummy}'], dummy=seq))
+    # garbage when combined with longer spec-sequences
+    # but this is unavoidable without introducing a whitelist
+    # of supported value types -- which would limit flexibility
+    eq_(["['one', 'two']", 'other'],
+        _format_iospecs(['{dummy}', 'other'], dummy=seq))
+
+
+def test_substitution_config():
+    # use a shim to avoid having to create an actual dataset
+    # the tested function only needs a `ds.config` to be a ConfigManager
+    from datalad import cfg
+
+    class dset:
+        config = cfg
+
+    # empty be default
+    eq_(_get_substitutions(dset), {})
+    # basic access
+    with patch_config({"datalad.run.substitutions.dummy": 'ork'}):
+        eq_(_get_substitutions(dset), dict(dummy='ork'))
+    # can report multi-value
+    with patch_config({"datalad.run.substitutions.dummy": ['a', 'b']}):
+        eq_(_get_substitutions(dset), dict(dummy=['a', 'b']))
+
+        # verify combo with iospec formatting
+        eq_(_format_iospecs(['{dummy}'],
+                            **_get_substitutions(dset)),
+            ['a', 'b'])
